import numpy as np

###############################################
####### creating the maze enviroment ##########
###############################################

class Maze(object):
    
    def __init__(self, n):
        self.current_state = (0, 0)
        self.size = n
        self.n_states = n**2
        self.n_actions = 4
        self.done = False

    def step(self, action):
        # returns the next state and the reward 
        # encoding of the actions: 0-down, 1-right, 2-up, 3-left.
        n = self.size
        if(action == 0): 
            self.current_state = (np.max((self.current_state[0] - 1, 0)), self.current_state[1])
        if(action == 1): #east
            self.current_state = (self.current_state[0], np.min((self.current_state[1] + 1, n-1)))
        if(action == 2): 
            self.current_state = (np.min((self.current_state[0] + 1, n-1)), self.current_state[1])
        if(action == 3): 
            self.current_state = (self.current_state[0], np.max((self.current_state[1] - 1, 0)))
            
        reward = -1
        if self.current_state == (0, n-1) or self.current_state == (n-1, 0):
            reward = 10
            self.done = True
        
        return self.current_state, reward, self.done
    
    def reset(self):
        self.current_state = (0, 0)
        self.done = False
    
    
def ind_to_cell(k, n):
    # converts k from 1 to n^2 type state encoding to 
    # (i, j) type state encoding
    i = int(k / n)
    j = int(k % n) 
    return i, j

def cell_to_ind(state, n):
    # converts (i, j) type state encoding to 
    # k from 1 to n^2 type state encoding
    i = state[0]
    j = state[1]
    return i * n + j



###############################################
########## policy of the agent ################
###############################################

def pick_action(state, Q_table, epsilon=0):
    # implements greedy policy and returns the action with max. Q-value (given the state).
    # note: when Q-table is filled with zeros, returns a random policy. 
    n = int(np.sqrt(Q_table.shape[0]))
    state_ind = cell_to_ind(state, n)
    action_scores = Q_table[state_ind]
    
    action = np.random.choice(np.where(action_scores == np.max(action_scores))[0])
    if np.random.random() < epsilon:
        action = np.random.choice(np.arange(Q_table.shape[1]))
    return action


###############################################
############ helper functions #################
###############################################

def calc_V(Q_table, epsilon=0):
    # averages over the actions on the Q-table and returns 
    # the corresponding V-table
    n = int(np.sqrt(Q_table.shape[0]))
    V_table = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            state_ind = cell_to_ind((i, j), n)
            V_table[n-1-i, j] = np.max(Q_table[state_ind]) * (1-epsilon) +  epsilon * np.mean(Q_table[state_ind])
    return V_table